/* $Id: KeyboardAppUi.cpp,v 1.22 2004/09/11 21:12:17 chrisf Exp $ */

/* Copyright (c) 2003 Chris Faherty */

/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <aknquerydialog.h>
#include <avkon.hrh>
#include <eikmenup.h>
#include <eikgted.h>
#include <apgwgnam.h>
#include <Keyboard.rsg>
#include "Keyboard.pan"
#include "KeyboardAppUi.h"
#include "KeyboardAppView.h"
#include "Keyboard.hrh"
#include "KeyboardAppModel.h"

static const TUid KUidKeyboardApp = { 0x101FEAE0 };
static const TUid KUidPhoneStandbyApp = { 0x100058B3 };

// ConstructL is called by the application framework
void CKeyboardAppUi::ConstructL()
{
    BaseConstructL();

    // Create view
    iAppView = CKeyboardAppView::NewL(ClientRect());
    AddToStackL(iAppView);

    // Create engine
    iAppModel = CKeyboardAppModel::NewL();
    iAppView->SetModel(iAppModel);
    iAppModel->SetView(iAppView);

/* temporarily take out */
    iCaptureKeyHandle = CEikonEnv::Static()->
        RootWin().CaptureKey(EKeyLeftShift, 0, 0);
    iCaptureKeyUpAndDownsHandle = CEikonEnv::Static()->
        RootWin().CaptureKeyUpAndDowns(EStdKeyLeftShift, 0, 0);
}

CKeyboardAppUi::CKeyboardAppUi() :
    iCaptureKeyHandle(0),
    iCaptureKeyUpAndDownsHandle(0),
    iCycledTask(0)
{
    CEikonEnv::Static()->SetSystem(true);
}

CKeyboardAppUi::~CKeyboardAppUi()
{
/* temporarily take out */
    CEikonEnv::Static()->
        RootWin().CancelCaptureKey(iCaptureKeyHandle);
    iCaptureKeyHandle = 0;
    CEikonEnv::Static()->
        RootWin().CancelCaptureKeyUpAndDowns(iCaptureKeyUpAndDownsHandle);
    iCaptureKeyUpAndDownsHandle = 0;

    delete iAppModel;
    iAppModel = NULL;

    if (iAppView) {
        RemoveFromStack(iAppView);
        delete iAppView;
        iAppView = NULL;
    }
}

// handle any menu commands
void CKeyboardAppUi::HandleCommandL(TInt aCommand)
{
    switch (aCommand) {
    case EAknSoftkeyExit:
        Exit();
        break;

    case EKeyboardCmdOnPT:
        iAppView->setKeyboardModel(POCKETOP_MODEL);
        iAppView->Init();
        iAppModel->Start();
        break;

    case EKeyboardCmdOnPWUS:
        iAppView->setKeyboardModel(THINKOUTSIDE_MODEL);
        iAppView->setKeyboardLayout(US_LAYOUT);
        iAppView->Init();
        iAppModel->Start();
        break;

    case EKeyboardCmdOnPWFR:
        iAppView->setKeyboardModel(THINKOUTSIDE_MODEL);
        iAppView->setKeyboardLayout(FR_LAYOUT);
        iAppView->Init();
        iAppModel->Start();
        break;

    case EKeyboardCmdOnPWDE:
        iAppView->setKeyboardModel(THINKOUTSIDE_MODEL);
        iAppView->setKeyboardLayout(DE_LAYOUT);
        iAppView->Init();
        iAppModel->Start();
        break;

    case EKeyboardCmdOnDB:
        iAppView->setKeyboardModel(DEBUG_MODEL);
        iAppView->Init(); /* debug, show ascified chars */
        iAppModel->Start();
        break;

    case EKeyboardCmdOff:
        iAppModel->Cancel();
        iAppModel->Stop();
        break;

    default:
        User::Panic(KPanicKeyboard, EKeyboardBasicUi);
        break;
    }
}

void CKeyboardAppUi::DynInitMenuPaneL(TInt aMenuId, CEikMenuPane * aMenuPane)
{
    if (aMenuId == R_KEYBOARD_MENU) {
        // Grey-out 'On' menu items if already connected
        aMenuPane->SetItemDimmed(EKeyboardCmdOnPT,
            iAppModel->Connected());
        aMenuPane->SetItemDimmed(EKeyboardCmdOnPWUS,
            iAppModel->Connected());
        aMenuPane->SetItemDimmed(EKeyboardCmdOnPWFR,
            iAppModel->Connected());
        aMenuPane->SetItemDimmed(EKeyboardCmdOnPWDE,
            iAppModel->Connected());
        aMenuPane->SetItemDimmed(EKeyboardCmdOnDB,
            iAppModel->Connected());

        // Grey-out 'Off' menu item if not connected
        aMenuPane->SetItemDimmed(EKeyboardCmdOff,
            !iAppModel->Connected());
    }
}

TKeyResponse CKeyboardAppUi::HandleKeyEventL(const TKeyEvent& aKeyEvent,
    TEventCode aType)
{
//    if (aKeyEvent.iScanCode != EStdKeyLeftShift) return EKeyWasNotConsumed;

    CCoeEnv *ccoeenv=CCoeEnv::Static();

    if ((aType == EEventKeyDown) && (aKeyEvent.iScanCode == EStdKeyLeftShift)) {
        /* do my task switch */
        TApaTaskList tasklist(ccoeenv->WsSession());
        TInt num=ccoeenv->WsSession().NumWindowGroups(0);
        if (++iCycledTask >= num) iCycledTask = 1;
        for (int i=0; i < num; i++) {
            TApaTask *task =
                new (ELeave)TApaTask(tasklist.FindByPos(iCycledTask));
            CleanupStack::PushL(task);
            CApaWindowGroupName *wgName=
                CApaWindowGroupName::NewLC(ccoeenv->WsSession(), task->WgId());
            if (!wgName->Hidden() &&
                (wgName->AppUid() != KUidKeyboardApp) &&
                (wgName->AppUid() != KUidPhoneStandbyApp)) {
                wgName->SetSystem(true);
                wgName->SetRespondsToShutdownEvent(false);
                task->BringToForeground();
                CleanupStack::PopAndDestroy(2);
                break;
            }
            CleanupStack::PopAndDestroy(2);
            if (++iCycledTask >= num) iCycledTask = 1;
        }
    }

    return EKeyWasNotConsumed;
}
