/* $Id: KeyboardAppView.cpp,v 1.38 2004/09/11 22:03:09 chrisf Exp $ */

/* Copyright (c) 2003 Chris Faherty */

/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <eiklabel.h>
#include <eikrted.h>
#include <eikenv.h>
#include <aknutils.h>
#include <aknedsts.h>
#include <apgtask.h>
#include <coeaui.h>
#include <barsread.h>           // for resource reader
#include <stringloader.h>
#include <Keyboard.rsg>
#include <aknglobalnote.h>
#include <string.h>
#include <stdio.h>

#include "KeyboardLayouts.h"
#include "KeyboardAppView.h"
#include "Keyboard.pan"

const unsigned char layout_offsets[] = {
    0x1b, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x3c, 0x3d, 0x3e, 0x3f, 0x38, 0x39, 0x32,
    0x18,    0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x44, 0x45, 0x46, 0x47, 0x56,    0x57,
    0x58,      0x03, 0x10, 0x2c, 0x2d, 0x2e, 0x2f, 0x4c, 0x4d, 0x4e, 0x48, 0x59,  0x3a,
    0x1a, 0x08,  0x23,  0x22, 0x17,                0x37, 0x24, 0x51, 0x52, 0x53,  0x50
};

// Standard Epoc construction sequence
CKeyboardAppView *CKeyboardAppView::NewL(const TRect &aRect)
{
    CKeyboardAppView *self = CKeyboardAppView::NewLC(aRect);
    CleanupStack::Pop(self);
    return self;
}

CKeyboardAppView *CKeyboardAppView::NewLC(const TRect &aRect)
{
    CKeyboardAppView *self = new (ELeave)CKeyboardAppView;
    CleanupStack::PushL(self);
    self->ConstructL(aRect);
    return self;
}

void CKeyboardAppView::SetModel(CKeyboardAppModel *aModel)
{
    iModel = aModel;
}

int CKeyboardAppView::getKeyboardModel()
{
    return keyboard_model;
}

int CKeyboardAppView::getKeyboardLayout()
{
    return keyboard_layout;
}

void CKeyboardAppView::setKeyboardModel(int model)
{
    keyboard_model= model;
}

void CKeyboardAppView::setKeyboardLayout(int layout)
{
    keyboard_layout= layout;
    if (keyboard_model == THINKOUTSIDE_MODEL) {
        switch (keyboard_layout) {
        case US_LAYOUT:
            initThinkoutsideLayout(layout_us_plain, layout_us_shifted,
                layout_us_func, layout_us_func2);
            break;
        case FR_LAYOUT:
            initThinkoutsideLayout(layout_fr_plain, layout_fr_shifted,
                layout_fr_func, layout_fr_func2);
            break;
        case DE_LAYOUT:
            initThinkoutsideLayout(layout_de_plain, layout_de_shifted,
                layout_de_func, layout_de_func2);
            break;
        }
    }
}

void CKeyboardAppView::ConstructL(const TRect &aRect)
{
    // Definitions
    const TPoint KLabelPosition = TPoint(10, 10);
    const TSize KLabelSize = TSize(150, 80);
    const TPoint KOutputPosition = TPoint(10, 90);
    const TSize  KOutputSize     = TSize(150, 40);

    CreateWindowL();
    TResourceReader reader;

    iCoeEnv->CreateResourceReaderLC(reader, R_ABOUT_LABEL);
    iStatusWindow = new (ELeave)CEikLabel();
    iStatusWindow->SetContainerWindowL(*this);
    iStatusWindow->ConstructFromResourceL(reader);
    CleanupStack::PopAndDestroy();  // Resource reader
    iStatusWindow->SetFocus(ETrue);
    iStatusWindow->SetExtent(KLabelPosition, KLabelSize); 
    iStatusWindow->SetFont(LatinPlain12());

    iCoeEnv->CreateResourceReaderLC(reader, R_TEST_RTEXTED);
    iOutputWindow = new (ELeave)CEikRichTextEditor();
    iOutputWindow->SetContainerWindowL(*this);
    iOutputWindow->ConstructFromResourceL(reader);
    CleanupStack::PopAndDestroy();  // Resource reader
    iOutputWindow->SetFocus(ETrue);
    iOutputWindow->SetExtent(KOutputPosition, KOutputSize); 

    SetRect(aRect);
    ActivateL();

    iTimer = CPeriodic::NewL(CActive::EPriorityStandard);
    iForceBreakTimer = CPeriodic::NewL(CActive::EPriorityStandard);
}

CKeyboardAppView::CKeyboardAppView()
{
    // defaults
    keyboard_model = THINKOUTSIDE_MODEL;
    setKeyboardLayout(US_LAYOUT);
}

CKeyboardAppView::~CKeyboardAppView()
{
    iTimer->Cancel();
    delete iTimer;
    iTimer = NULL;
    iForceBreakTimer->Cancel();
    ForceBreakTick(this); /* better do a forcebreak to be safe */
    delete iForceBreakTimer;
    iForceBreakTimer = NULL;
    delete iStatusWindow;
    iStatusWindow = NULL;
    delete iOutputWindow;
    iOutputWindow = NULL;
}

void CKeyboardAppView::initThinkoutsideLayout(
    const kchar * layout_plain,
    const kchar * layout_shifted,
    const kchar * layout_func,
    const kchar * layout_func2)
{
    for (unsigned int i=0; i < sizeof(layout_offsets); i++) {
        int offset=layout_offsets[i];
        thinkoutside_plain[offset] = layout_plain[i];
        thinkoutside_shifted[offset] = layout_shifted[i];
        thinkoutside_func[offset] = layout_func[i];
        thinkoutside_func2[offset] = layout_func2[i];
    }
}

void CKeyboardAppView::Init()
{
    shift = capslock = ctrl = alt = func = pt_shift = num_shift = false;
    thinkoutside_state = 0;
    thinkoutside_isirdaframe = false;
    lastraw = 0;
    lastscode = -1;
    iOutputWindow->Text()->Reset();
    iOutputWindow->HandleTextChangedL();
    iOutputWindow->SetCursorPosL(0, EFalse);
    iTimer->Cancel();
    iForceBreakTimer->Cancel();
}

// Draw this application's view to the screen
void CKeyboardAppView::Draw(const TRect & aRect) const
{
    // Clear the screen
    CWindowGc & gc = SystemGc();
    gc.Clear(aRect);
}

TInt CKeyboardAppView::CountComponentControls() const
{
    return 2;
}

CCoeControl *CKeyboardAppView::ComponentControl(TInt aIndex) const
{

    switch (aIndex) {
    case 0:
        return iStatusWindow;
    case 1:
        return iOutputWindow;

    default:
        User::Panic(KPanicKeyboard, EKeyboardAppView);
        return 0;
    }
}

TKeyResponse CKeyboardAppView::OfferKeyEventL(const TKeyEvent &aKeyEvent,
    TEventCode aType)
{
    return iOutputWindow->OfferKeyEventL(aKeyEvent, aType);
}

void CKeyboardAppView::Incoming(TDesC8 &buf)
{
    int i;
    switch (keyboard_model) {
    case POCKETOP_MODEL:
        for (i=0; i < buf.Length(); i++) pocketop(buf[i], false);
        break;
    case THINKOUTSIDE_MODEL:
        for (i=0; i < buf.Length(); i++) thinkoutside(buf[i], false);
        break;
    case DEBUG_MODEL:
        for (i=0; i < buf.Length(); i++) debug(buf[i], false);
        break;
    }
    User::ResetInactivityTime(); /* keep backlight on */
    tick = 0;
    iTimer->Cancel();
    /* 10 second periodic timer */
    iTimer->Start(10*1000*1000, 10*1000*1000, TCallBack(Tick, this));
}

TInt CKeyboardAppView::Tick(TAny *aObject)
{
    CKeyboardAppView *taui=(CKeyboardAppView *)aObject;
    if (++(taui->tick) < 3) { /* 30 seconds */
        User::ResetInactivityTime(); /* keep backlight on */
        return 1;
    }
    return 0;
}

TInt CKeyboardAppView::ForceBreakTick(TAny *aObject)
{
    CKeyboardAppView *taui=(CKeyboardAppView *)aObject;
    switch (taui->keyboard_model) {
    case POCKETOP_MODEL:
        taui->pocketop(0, true);
        break;
    case THINKOUTSIDE_MODEL:
        taui->thinkoutside(0, true);
        break;
    case DEBUG_MODEL:
        taui->debug(0, true);
        break;
    }
    return 0;
}

void CKeyboardAppView::pocketop(unsigned char raw, TBool forcebreak)
{
    kchar cook;
    unsigned int modifier=EModifierAutorepeatable;
    TRawEvent aEvent;

    CCoeEnv *ccoeenv=CCoeEnv::Static();

    if (forcebreak) {
        raw = (unsigned char)(lastraw|0x80); /* force break code */
    }

    lastraw = raw;
    cook = (unsigned char)(raw&0x7f);
    if (cook >= sizeof(pockettop_lut)) return;
    TInt code=pockettop_lut[cook];
    TInt scode=0;
    switch (code) {
    case KB_ENTER: code = EKeyEnter; break;
    case KB_BACKSPACE: code = EKeyBackspace; break;
    case KB_UP: code = EKeyUpArrow; break;
    case KB_DOWN: code = EKeyDownArrow; break;
    case KB_LEFT: code = EKeyLeftArrow; break;
    case KB_RIGHT: code = EKeyRightArrow; break;
    case KB_TAB: code = EKeyTab; break;
    case KB_ESC: code = EKeyEscape; break;
    case KB_DELETE: code = EKeyDelete; break;
    case KB_COMMAND:
        /* haven't figured out what I want yet */
        break;
    case KB_DATE:
        code = 0; /* custom1 */
        scode = EStdKeyDevice0;
        break;
    case KB_ADDRESS:
        code = 0; /* custom2 */
        scode = EStdKeyDevice1;
        break;
    case KB_TODO:
        code = 0; /* dial press */
        scode = EStdKeyDevice3;
        break;
    case KB_MEMO:
        code = 0;
        scode = EStdKeyApplication0;
        break;
    case KB_LSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_RSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_SHORTCUT:
        pt_shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_NUMLOCK:
        num_shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_CAPLOCK:
        if (!(raw&0x80)) capslock = !capslock; /* toggle */
        code = 0;
        break;
    case KB_CTL:
        ctrl = (bool)!(raw&0x80);
        code = EKeyLeftCtrl;
        break;
    case KB_ALT:
        alt = (bool)!(raw&0x80);
        code = EKeyLeftAlt;
        break;
    case KB_FUNC:
        if (num_shift) {
            code = '0';
            break;
        }
        func = (bool)!(raw&0x80);
        code = 0;
        break;
    }

    if (raw&0x80) { /* high-bit on is break code */
        /* send keyup event */
        aEvent.Set(TRawEvent::EKeyUp, scode);
        ccoeenv->WsSession().SimulateRawEvent(aEvent);
        return;
    }
    if (!code && !scode) return;

    if (shift) {
        /* other chars */
        switch (code) {
        case '-': code = '_'; break;
        case ';': code = ':'; break;
        case ',': code = '<'; break;
        case '.': code = '>'; break;
        }
    }

    /* punct keys */
    if (pt_shift) {
        switch (code) {
        case 'q': code = '!'; break;
        case 'w': code = '@'; break;
        case 'e': code = '#'; break;
        case 'r': code = '$'; break;
        case 't': code = '%'; break;
        case 'y': code = '^'; break;
        case 'u': code = '&'; break;
        case 'i': code = '*'; break;
        case 'o': code = '('; break;
        case 'p': code = ')'; break;
        case '-': code = '+'; break;
        case 'a': code = '`'; break;
        case 's': code = '~'; break;
        case 'd': code = '/'; break;
        case 'f': code = '?'; break;
        case 'g': code = '['; break;
        case 'h': code = ']'; break;
        case 'j': code = '{'; break;
        case 'k': code = '}'; break;
        case 'l': code = '\\'; break;
        case ';': code = '|'; break;
        case '\'': code = '='; break;
        }
    }

    /* blue Function key */
    if (func) {
        switch (code) {
        case EKeyTab: code = EKeyEscape; break;
        }
    }

    /* numlock keys */
    if (num_shift) {
        switch (code) {
        case 'q': code = '1'; break;
        case 'w': code = '2'; break;
        case 'e': code = '3'; break;
        case 'r': code = '4'; break;
        case 't': code = '5'; break;
        case 'y': code = '6'; break;
        case 'u': code = '7'; break;
        case 'i': code = '8'; break;
        case 'o': code = '9'; break;
        case 'p': code = '0'; break;
        case '-': code = '/'; break;
        case 'j': code = '4'; break;
        case 'k': code = '5'; break;
        case 'l': code = '6'; break;
        case ';': code = '*'; break;
        case 'm': code = '1'; break;
        case ',': code = '2'; break;
        case '.': code = '3'; break;
        case EKeyUpArrow: code = '-'; break;
        case EKeyLeftArrow: code = '.'; break;
        case EKeyDownArrow: code = '+'; break;
        case EKeyRightArrow: code = '='; break;
        }
    }

    /* make it into control chars if needed */
    if (ctrl) {
        if ((code >= 'a') && (code <= 'z')) code = code-'a'+1;
        if ((code >= '[') && (code <= '_')) code = code-'A';
        if ((code >= '2') && (code <= '7')) code = code-('2'-(char)26);
        modifier |= EModifierCtrl;
    }

    /* I still have to manually shift characters, even with mod */
    if (shift || capslock) {
        if ((code >= 'a') && (code <= 'z')) code -= ('a'-'A');
    }

    /* send keydown event */
    aEvent.Set(TRawEvent::EKeyDown, scode);
    ccoeenv->WsSession().SimulateRawEvent(aEvent);

    /* send keypress event */
    TKeyEvent aKeyEvent;
    aKeyEvent.iCode = code;
    /* trying to autorepeat a code=0 crashes */
    if (!code) 
        aKeyEvent.iModifiers = 0;
    else
        aKeyEvent.iModifiers = modifier;

    aKeyEvent.iRepeats = 0;
    aKeyEvent.iScanCode = scode;
    ccoeenv->WsSession().SimulateKeyEvent(aKeyEvent);

    /*
     * 10 second timer.  When it expires I will force a keyup.  The timer
     * is reset whenever we receive a repeat signal from the keyboard.  This
     * timer is 10 seconds because the Pocketop keyboard doesn't send any
     * keep-alive signal.  So I will make 10 seconds be the longest key hold
     * to keep a communications break from turning into an endless repeat.
     */
    iForceBreakTimer->Cancel();
    iForceBreakTimer->Start(10*1000*1000, 10*1000*1000,
        TCallBack(ForceBreakTick, this));
}

void CKeyboardAppView::thinkoutside(unsigned char raw, TBool forcebreak)
{
    kchar cook;
    unsigned int modifier=EModifierAutorepeatable;
    TRawEvent aEvent;

    CCoeEnv *ccoeenv=CCoeEnv::Static();

    if (forcebreak) {
        if (lastscode != -1) {
            /* send keyup event */
            aEvent.Set(TRawEvent::EKeyUp, lastscode);
            ccoeenv->WsSession().SimulateRawEvent(aEvent);
            lastscode = -1;
        }
        return;
    }

    /*
     * ThinkOutside IrDA protocol.  0xffc0 scancode 3bytesofstuff(prob crc).
     *
     * I'm going to wait for 0xffc0 then interpret the scancode and ignore
     * until the next 0xffc0.
     *
     * It's come to my attention that there are some Palm Wireless Keyboards
     * which do not send IrDA framing.  Therefore I have adjusted this to
     * handle both cases.
     */
    switch (thinkoutside_state) {
    case 0:
        /* IrDA frame start */
        if (raw == 0xff) {
            thinkoutside_state = 1;
            thinkoutside_isirdaframe = true;
        } else {
            /* otherwise it's a scancode */
            thinkoutside_isirdaframe = false;
            thinkoutside_state = 2;
            goto thinkoutside_state2; /* Weeeee! */
        }
        return;
    case 1:
        if (raw == 0xc0) thinkoutside_state = 2;
        else thinkoutside_state = 0;
        return;
    case 2:
thinkoutside_state2:
        if (thinkoutside_isirdaframe) thinkoutside_state = 3;
        else thinkoutside_state = 0;
        thinkoutside_isirdaframe = false; /* reset our flag */
        if ((raw == 0x5c)) { /* repeat indicator */
            /* reset the timer so it doesn't forcebreak on us */
            iForceBreakTimer->Cancel();
            iForceBreakTimer->Start(1*1000*1000, 1*1000*1000,
                TCallBack(ForceBreakTick, this));
            return;
        }
        break;
    case 3:
        /*
         * I'm waiting for a 0xc1 which terminates an IrDA frame.  That way
         * I can assume that a 0xff will indicate an IrDA frame next time.
         * If there is a 0xc1 within any CRC then I am screwed.  I didn't
         * manually check them all yet.
         */
        if (raw == 0xc1) thinkoutside_state = 0;
        return;
    }

    if ((raw&0x80) && (lastscode != -1)) {
        /* send keyup event */
        aEvent.Set(TRawEvent::EKeyUp, lastscode);
        ccoeenv->WsSession().SimulateRawEvent(aEvent);
        lastscode = -1;
    }

    cook = (unsigned char)(raw&0x7f);
    if (cook >= THINKOUTSIDE_MAX) return;
    TInt code=thinkoutside_plain[cook];
    TInt scode=0;
    switch (code) {
    case KB_COMMAND:
        /* haven't figured out what I want yet */
        break;
    case KB_LSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_RSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_FUNC2:
        pt_shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_CAPLOCK:
        if (!(raw&0x80)) capslock = !capslock; /* toggle */
        code = 0;
        break;
    case KB_CTL:
        ctrl = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_ALT:
        alt = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_FUNC:
        func = (bool)!(raw&0x80);
        code = 0;
        break;
    }

    if (raw&0x80) return; /* high-bit on is break code */
    if (!code && !scode) return;

    if (shift) {
        code = thinkoutside_shifted[cook];
        //modifier |= EModifierShift; /* doesn't work right on 3650 */
    }

    if (capslock) {
        /* for capslock -- only shift alpha characters */
        if ((code >= 'a') && (code <= 'z')) code = thinkoutside_shifted[cook];
        //modifier |= EModifierShift; /* doesn't work right on 3650 */
    }

    if (pt_shift) {
        code = thinkoutside_func2[cook];    
        modifier |= EModifierRightFunc;
    }
    
    if (alt) {
        modifier |= EModifierAlt;
    }

    if (func) {
        code = thinkoutside_func[cook];
        modifier |= EModifierLeftFunc;
    }

    switch (code) {
        case KB_ENTER: code = EKeyEnter; break;
        case KB_BACKSPACE: code = EKeyBackspace; break;
        case KB_UP: code = EKeyUpArrow; break;
        case KB_DOWN: code = EKeyDownArrow; break;
        case KB_LEFT: code = EKeyLeftArrow; break;
        case KB_RIGHT: code = EKeyRightArrow; break;
        case KB_TAB: code = EKeyTab; break;
        case KB_ESC: code = EKeyEscape; break;
        case KB_DELETE: code = EKeyDelete; break;
        case KB_DATE:
            code = 0; /* custom1 */
            scode = EStdKeyDevice0;
            break;
        case KB_ADDRESS:
            code = 0; /* custom2 */
            scode = EStdKeyDevice1;
            break;
        case KB_TODO:
            code = 0; /* dial press */
            scode = EStdKeyDevice3;
            break;
        case KB_MEMO:
            code = 0;
            scode = EStdKeyApplication0;
            break;
        }

    /* make it into control chars if needed */
    if (ctrl) {
        if ((code >= 'a') && (code <= 'z')) code = code-'a'+1;
        if ((code >= '[') && (code <= '_')) code = code-'A';
        if ((code >= '2') && (code <= '7')) code = code-('2'-(char)26);
        modifier |= EModifierCtrl;
    }

    /* send keydown event */
    aEvent.Set(TRawEvent::EKeyDown, scode);
    ccoeenv->WsSession().SimulateRawEvent(aEvent);
    lastscode = scode;

    /* send keypress event */
    TKeyEvent aKeyEvent;
    aKeyEvent.iCode = code;
    /* trying to autorepeat a code=0 crashes */
    if (!code) 
        aKeyEvent.iModifiers = 0;
    else
        aKeyEvent.iModifiers = modifier;
    
    aKeyEvent.iRepeats = 0;
    aKeyEvent.iScanCode = scode;
    ccoeenv->WsSession().SimulateKeyEvent(aKeyEvent);

    /*
     * 1 second timer.  When it expires I will force a keyup.  The timer
     * is reset whenever we receive a repeat signal from the keyboard.
     */
    iForceBreakTimer->Cancel();
    iForceBreakTimer->Start(1*1000*1000, 1*1000*1000,
        TCallBack(ForceBreakTick, this));
}

void CKeyboardAppView::debug(unsigned char raw, TBool forcebreak)
{
    TKeyEvent aKeyEvent;
    unsigned char hi, lo;

    CCoeEnv *ccoeenv=CCoeEnv::Static();

    aKeyEvent.iModifiers = 0;
    aKeyEvent.iRepeats = 0;
    aKeyEvent.iScanCode = 0;

    /* send an ascified version of the character */
    hi = raw; hi >>= 4; hi += '0';
    if (hi > '9') hi += ('a'-':');
    lo = raw; lo &= 0xf; lo += '0';
    if (lo > '9') lo += ('a'-':');
    aKeyEvent.iCode = hi;
    ccoeenv->WsSession().SimulateKeyEvent(aKeyEvent);
    aKeyEvent.iCode = lo;
    ccoeenv->WsSession().SimulateKeyEvent(aKeyEvent);
}
